;===========================================================================================
; RotBit.asm  GoAsm assembler
; Rotation de bits pour texte et bitmap
; Programm par AsmGges France
;===========================================================================================
; compile.bat:
; ..\..\bin\GoRC /r, res\rotbit.rc
; ..\..\bin\GoAsm rotbit.asm
; ..\..\bin\GoLink rotbit.obj res\rotbit.res user32.dll kernel32.dll gdi32.dll
;===========================================================================================

; Inclusion des macros/quates/structures ncessaires au programme
; ----------------------------------------------------------------
  #include "..\..\inc\Windows.inc"
  #include "..\..\inc\all_api.inc"

  ID_FRANCE = 17

  ID_BITMAP = 100
  ID_EXIT   = 200

;===========================================================================================
; Dbut du code programme
; -----------------------
.code

start:
    ; Obtenir le handle du programme
    ; ------------------------------
    invoke GetModuleHandle,0
    mov [wc.hInstance],eax

    ; Chargement d'une icne spcifique
    ; ---------------------------------
    invoke LoadIcon,[wc.hInstance],ID_FRANCE
    mov [wc.hIcon],eax

    ; Chargement d'un curseur systme
    ; -------------------------------
    invoke LoadCursor,0,IDC_ARROW
    mov [wc.hCursor],eax

    ; Obtenir la largeur en pixels de l'cran
    ; ---------------------------------------
    invoke GetSystemMetrics,SM_CXSCREEN
    sub eax,[WinW]
    shr eax,1
    mov [WinX],eax

    ; Valider la structure WNDCLASS
    ; -----------------------------
    invoke RegisterClass,addr wc

    ; Cration de la fenetre de l'application
    ; ---------------------------------------
    invoke CreateWindowEx,0,addr ClassName,addr Titre,WS_CAPTION+WS_POPUP+WS_SYSMENU \
                          +WS_MINIMIZEBOX,[WinX],[WinY],[WinW],[WinH],0,0,[wc.hInstance],0
    mov [hWnd],eax

    ; Affiche et rafraichit la fenetre
    ; --------------------------------
    invoke ShowWindow,[hWnd],SW_SHOWDEFAULT
    invoke UpdateWindow,[hWnd]

 ; Boucle des messages
 ; -------------------
 MessageLoop:
    invoke GetMessage,addr msg,NULL,0,0
    or eax,eax
    jz >ExitProgram
    invoke TranslateMessage,addr msg
    invoke DispatchMessage,addr msg
    jmp <MessageLoop

 ExitProgram:
    ; Fermeture de l'application
    ; --------------------------
    invoke ExitProcess,[msg.wParam]

;===========================================================================================
; Procdure principale Windows
;-------------------------------------------------------------------------------------------
WindowProc frame @hWnd, @uMsg, @wParam, @lParam
    uses ebx,esi,edi
    local .DestDC:D 
    local .SrcDC:D
    local .hBmp1:D
    local .hBmp2:D

    cmp D[@uMsg],WM_CLOSE
    jne >.wmcreate
        ; Boite  message fin d'application
        ; ---------------------------------
        invoke MessageBox,[@hWnd],addr MsgEnd,addr ClassName,MB_OK+MB_ICONINFORMATION

        ; Demande de fin d'application
        ; ----------------------------
        invoke PostQuitMessage,0
        jmp >>.exit

    .wmcreate:
    cmp D[@uMsg],WM_CREATE
    jne >.wmpaint
        ; Dfinir une fonte de texte
        ; --------------------------
        invoke CreateFont,10,0,0,0,FW_NORMAL,0,0,0,ANSI_CHARSET,OUT_DEFAULT_PRECIS,\
                        CLIP_DEFAULT_PRECIS,DEFAULT_QUALITY,DEFAULT_PITCH+FF_SWISS,\
                                                                       addr FontMS
        mov [hFontMS],eax                      ;fonte handle

        ; Construit tous les controles
        ; ----------------------------
        invoke CreateControls,[@hWnd]
        jmp >>.exit

    .wmpaint:
    cmp D[@uMsg],WM_PAINT
    jne >>.wmcommand
        invoke GetClientRect,[@hWnd],addr rect1
        cmp D[rect1.right],0 
        je >>.exit
            cmp D[rect1.bottom],0
            je >>.exit
                invoke BeginPaint,[@hWnd],addr ps

                ; Cration d'un dispositif de contexte mmoire pour la destination
                ; ----------------------------------------------------------------
                invoke CreateCompatibleDC,[ps.hdc]
                mov [.DestDC],eax

                ; Cration du bitmap de destination
                ; ---------------------------------
                invoke CreateCompatibleBitmap,[ps.hdc],[rect1.right],[rect1.bottom]
                mov [.hBmp1],eax

                ; L'inclure dans le dispositif de contexte de destination
                ; -------------------------------------------------------
                invoke SelectObject,[.DestDC],[.hBmp1]
                push eax

                ; Fixer la couleur du fond de la destination
                ; ------------------------------------------
                invoke GetStockObject,WHITE_BRUSH
                invoke FillRect,[.DestDC],addr rect1,eax
    
                ; Cration d'un dispositif de contexte mmoire pour la source
                ; -----------------------------------------------------------
                invoke CreateCompatibleDC,[ps.hdc]
                mov [.SrcDC],eax

                ; Obtenir la taille du texte  dessiner
                ; -------------------------------------
                mov D[rect2.left],0
                mov D[rect2.top],0
                mov D[rect2.right],0
                mov D[rect2.bottom],0
                invoke DrawText,[.SrcDC],addr Texte,-1,addr rect2,DT_CALCRECT \
                                                +DT_SINGLELINE+DT_LEFT+DT_TOP

                ; Cration d'un bitmap pour le texte
                ; ----------------------------------
                invoke CreateCompatibleBitmap,[ps.hdc],[rect2.right],[rect2.bottom]
                mov [.hBmp2],eax

                ; L'inclure dans le dispositif de contexte de source
                ; --------------------------------------------------
                invoke SelectObject,[.SrcDC],[.hBmp2]
                push eax

                ; Dessiner le texte
                ; -----------------
                invoke DrawText,[.SrcDC],addr Texte,-1,addr rect2,DT_SINGLELINE \
                                                                +DT_LEFT+DT_TOP

                ; Rotation du bitmap
                ; ------------------
                invoke Rotate,[.hBmp1],[.hBmp2],50,20,0
                invoke Rotate,[.hBmp1],[.hBmp2],100,20,1
                invoke Rotate,[.hBmp1],[.hBmp2],150,20,2
                invoke Rotate,[.hBmp1],[.hBmp2],200,20,3

                ; Slection de l'ancien bitmap
                ; ----------------------------
                pop eax
                invoke SelectObject,[.SrcDC],eax

                ; Supprimer le bitmap source cre
                ; -------------------------------
                invoke DeleteObject,eax

                ; Charger un bitmap depuis une ressource
                ; --------------------------------------
                invoke LoadBitmap,[wc.hInstance],ID_BITMAP
                mov [.hBmp2],eax

                ; L'inclure dans le dispositif de contexte de source
                ; --------------------------------------------------
                invoke SelectObject,[.SrcDC],[.hBmp2]
                push eax

                ; Rotation du bitmap
                ; ------------------
                invoke Rotate,[.hBmp1],[.hBmp2],50,100,0
                invoke Rotate,[.hBmp1],[.hBmp2],100,100,1
                invoke Rotate,[.hBmp1],[.hBmp2],150,100,2
                invoke Rotate,[.hBmp1],[.hBmp2],200,100,3

                ; Slection de l'ancien bitmap source
                ; -----------------------------------
                pop eax
                invoke SelectObject,[.SrcDC],eax

                ; Supprimer le bitmap source cre
                ; -------------------------------
                invoke DeleteObject,eax

                ; Supprimer le dispositif de contexte de source
                ; ---------------------------------------------
                invoke DeleteDC,[.SrcDC]

                ; Copier le bitmap destination dans la fentre bitmap
                ; ---------------------------------------------------
                invoke BitBlt,[ps.hdc],[rect1.left],[rect1.top],[rect1.right],\
                                        [rect1.bottom],[.DestDC],0,0,SRCCOPY

                ; Slection de l'ancien bitmap destination
                ; ----------------------------------------
                pop eax
                invoke SelectObject,[.DestDC],eax

                ; Supprimer le bitmap destination cre
                ; ------------------------------------
                invoke DeleteObject,eax

                ; Supprimer le dispositif de contexte destination
                ; -----------------------------------------------
                invoke DeleteDC,[.DestDC]
                invoke EndPaint,[@hWnd],addr ps
                jmp >.exit

    .wmcommand:
    cmp D[@uMsg],WM_COMMAND
    jne >.defwindowproc
        mov eax,[@wParam]
        mov edx,eax
        shr edx,16
        and eax,0FFFFh
        cmp edx,BN_CLICKED
        jne >.exit
            cmp eax,ID_EXIT
            jne >.exit
                ; Demande de fin d'application
                ; ----------------------------
                invoke SendMessage,[@hWnd],WM_CLOSE,0,0
                jmp >.exit

    .defwindowproc:
        ; Traitement par dfaut des autres messages Windows
        ; -------------------------------------------------
        invoke DefWindowProc,[@hWnd],[@uMsg],[@wParam],[@lParam]
        ret

    .exit:
    mov eax,0
    ret
endf

;===========================================================================================
Rotate frame @hBmpDest, @hBmpSrc ,@x ,@y, @nRotate
    uses esi,edi
    local .nbitsd:D
    local .hmemd:D
    local .nbitss:D
    local .hmems:D

    ; Obtenir des infos sur le bitmap destination
    ; -------------------------------------------
    invoke GetObject,[@hBmpDest],sizeof BITMAP,addr bmd
    mov eax,[bmd.bmWidthBytes]
    mov edx,[bmd.bmHeight]
    mul edx
    mov [.nbitsd],eax

    ; Allouer de la mmoire pour les bits du bitmap destination
    ; ---------------------------------------------------------
    invoke GlobalAlloc,GMEM_FIXED,[.nbitsd]
    mov [.hmemd],eax

    ; Obtenir des infos sur les bits du bitmap destination
    ; ----------------------------------------------------
    invoke GetBitmapBits,[@hBmpDest],[.nbitsd],[.hmemd]

    ; Obtenir des infos sur le bitmap source
    ; --------------------------------------
    invoke GetObject,[@hBmpSrc],sizeof BITMAP,addr bms
    mov eax,[bms.bmWidthBytes]
    mov edx,[bms.bmHeight]
    mul edx
    mov [.nbitss],eax

    ; Allouer de la mmoire pour les bits du bitmap
    ; ---------------------------------------------
    invoke GlobalAlloc,GMEM_FIXED,[.nbitss]
    mov [.hmems],eax

    ; Obtenir les bits source du bitmap
    ; ---------------------------------
    invoke GetBitmapBits,[@hBmpSrc],[.nbitss],[.hmems]

    ; Copier les pixels un par un
    ; ---------------------------
    mov edx,0

    .bmpheight:
    cmp edx,[bms.bmHeight]
    jae >>.copiebits
        mov ecx,0

        .bmpwidth:
        cmp ecx,[bms.bmWidth]
        jae >>.endbmpwidth
            push ecx,edx
            mov esi,[.hmems]
            push edx
            mov eax,[bms.bmWidthBytes]
            mul edx
            add esi,eax
            movzx eax,B[bms.bmBitsPixel]
            shr eax,3
            mul ecx
            add esi,eax
            pop edx
            mov eax,[@nRotate]

            cmp eax,1
            jne >
                ; Rotation 90 degres
                ; ------------------
                sub edx,[bms.bmHeight]
                neg edx
                xchg ecx,edx
            :
            cmp eax,2
            jne >
                ; Rotation 180 degres
                ; -------------------
                sub edx,[bms.bmHeight]
                neg edx
                sub ecx,[bms.bmWidth]
                neg ecx
            :
            cmp eax,3
            jne >.caladr
                ; Rotation 270 degres
                ; -------------------
                sub ecx,[bms.bmWidth]
                neg ecx
                xchg ecx,edx

            .caladr:
            ; Ajouter les offsets de destination
            ; ----------------------------------
            add ecx,[@x]
            add edx,[@y]
            cmp ecx,[bmd.bmWidth]
            jae > 
                cmp edx,[bmd.bmHeight]
                jae >
                    ; Calculer l'adresse de destination
                    ; ---------------------------------
                    mov edi,[.hmemd]
                    mov eax,[bmd.bmWidthBytes]
                    mul edx
                    add edi,eax
                    movzx eax,B[bmd.bmBitsPixel]
                    shr eax,3
                    xchg eax,ecx
                    mul ecx
                    add edi,eax

                    ; Copier les octets
                    ; -----------------
                    rep movsb
                  
            :
            pop edx,ecx
            inc ecx
        jmp <.bmpwidth

        .endbmpwidth:
        inc edx
    jmp <.bmpheight

    .copiebits:
    ; Copier les bits du bitmap destination
    ; -------------------------------------
    invoke SetBitmapBits,[@hBmpDest],[.nbitsd],[.hmemd]

    ; Librer la mmoire alloue
    ; --------------------------
    invoke GlobalFree,[.hmems]
    invoke GlobalFree,[.hmemd]

    ret
endf

;===========================================================================================
CreateControls frame @hWnd

    ; Cration de la fentre-enfant PUSHBUTTON Exit
    ; ---------------------------------------------
    invoke Dlg,addr Button,addr TxtExit,[@hWnd],224,148,44,20,ID_EXIT,BS_PUSHBUTTON,0

    ret
endf

;===========================================================================================
; Cration d'une fentre-enfant control class
;-------------------------------------------------------------------------------------------
; @Class    = adresse du nom de control class
; @AdrText  = adresse du control text
; @hParent  = handle de la fentre parent
; @X        = x-coordonne fentre-enfant
; @Y        = y-coordonne fentre-enfant
; @Xw       = largeur en pixels fentre-enfant
; @Yh       = hauteur en pixels fentre-enfant
; @Id       = identificateur du controle forme ID_XXX
; @Style    = complment window style de WS_CHILD+WS_VISIBLE
; @ExStyle  = window style tendu
;-------------------------------------------------------------------------------------------
; => eax = handle de la fentre-enfant
;-------------------------------------------------------------------------------------------
Dlg frame @Class, @AdrText, @hParent, @X, @Y, @Xw, @Yh, @Id, @Style, @ExStyle

    mov eax,WS_CHILD+WS_VISIBLE
    add eax,[@Style]
    invoke CreateWindowEx,[@ExStyle],[@Class],[@AdrText],eax,[@X],[@Y],[@Xw],[@Yh],\
                          [@hParent],[@Id],[wc.hInstance],NULL
    mov ebx,eax

    ; Valider la fonte standard MS Sans Serif pour le control
    ; -------------------------------------------------------
    invoke SendMessage,ebx,WM_SETFONT,[hFontMS],NULL
    mov eax,ebx

    ret
endf

;===========================================================================================
; Dfinitions des datas chaines / variables initialises ou non initialises / structures
;-------------------------------------------------------------------------------------------
.data

  Titre       db ' Rotation de bits',0
  ClassName   db ' AsmGges Win32',0
  MsgEnd      db ' GoAsm AsmGges France @2006  ',0

  Texte       db 'GoAsm',0
  FontMS      db 'MS Sans Serif',0

  Button      db 'BUTTON',0
  TxtExit     db 'Exit',0

  align 4

  WinX        dd 0
  WinY        dd 100
  WinW        dd 292
  WinH        dd 212

  hWnd        dd ?
  hFontMS     dd ?

  wc          WNDCLASS    < 0,addr WindowProc,0,0,NULL,NULL,NULL,\
                            COLOR_BTNFACE+1,NULL,addr ClassName >
  msg         MSG         <>
  ps          PAINTSTRUCT <>
  rect1       RECT        <>
  rect2       RECT        <>
  bmd         BITMAP      <>
  bms         BITMAP      <>

;===========================================================================================
